import functools
import hashlib
import shutil
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import cast

import graphql
import pydantic
from graphql.utilities import value_from_ast_untyped

from iron_gql.util import capitalize_first


@dataclass(kw_only=True)
class Statement:
    raw_text: str
    file: Path
    lineno: int

    @property
    def location(self) -> str:
        return f"{self.file}:{self.lineno}"

    @property
    def clean_text(self) -> str:
        return textwrap.dedent(self.raw_text).strip()

    @property
    def hash_str(self) -> str:
        return hashlib.md5(self.clean_text.encode(), usedforsecurity=False).hexdigest()


@dataclass(kw_only=True)
class Query:
    stmt: Statement
    doc: graphql.DocumentNode
    schema: graphql.GraphQLSchema
    fragments: dict[str, graphql.FragmentDefinitionNode]
    exec_source: str

    @functools.cached_property
    def operation_def(self) -> graphql.OperationDefinitionNode:
        for op in self.doc.definitions:
            if isinstance(op, graphql.OperationDefinitionNode):
                return op
        msg = "No operation definition found in the document"
        raise ValueError(msg)

    @property
    def name(self) -> str:
        if self.operation_def.name:
            return self.operation_def.name.value
        return f"query{capitalize_first(self.stmt.hash_str)}"

    @property
    def variables(self) -> list["GQLVar"]:
        return [
            _parse_var(var_def, schema=self.schema, context=self.stmt.location)
            for var_def in self.operation_def.variable_definitions
        ]

    @property
    def root_type(self) -> graphql.GraphQLObjectType:
        root_type = self.schema.get_root_type(self.operation_def.operation)

        if not root_type:
            msg = f"{self.operation_def.operation} is not defined in the schema"
            raise ValueError(msg)
        return root_type

    @property
    def selection_set(self) -> list["GQLVar"]:
        ctx = ParseContext(
            schema=self.schema,
            fragments=self.fragments,
            location=self.stmt.location,
        )
        return _parse_selection_set(
            self.operation_def.selection_set, self.root_type, ctx
        )


@dataclass(kw_only=True)
class GQLVar:
    name: str
    type: "GQLType"
    parent_type: graphql.GraphQLNamedType | None


@dataclass(kw_only=True)
class GQLType:
    not_null: bool
    default_value: graphql.UndefinedType | Any = graphql.Undefined


@dataclass(kw_only=True)
class GQLSingularType(GQLType):
    type: graphql.GraphQLNamedType


@dataclass(kw_only=True)
class GQLObjectType(GQLType):
    type: graphql.GraphQLNamedType
    selection: list[GQLVar]


@dataclass(kw_only=True)
class GQLListType(GQLType):
    type: GQLType


@dataclass(kw_only=True)
class ParseContext:
    schema: graphql.GraphQLSchema
    fragments: dict[str, graphql.FragmentDefinitionNode]
    location: str = ""


def _error_with_context(message: str, context: str = "") -> str:
    return f"{message} in {context}" if context else message


def _parse_type_node(
    type_node: graphql.TypeNode,
    *,
    schema: graphql.GraphQLSchema,
    not_null: bool = False,
    context: str = "",
) -> GQLType:
    match type_node:
        case graphql.NamedTypeNode(name=name):
            gql_type = schema.get_type(name.value)
            if not gql_type:
                raise ValueError(
                    _error_with_context(f"Unknown type: {name.value}", context)
                )
            return GQLSingularType(type=gql_type, not_null=not_null)
        case graphql.ListTypeNode(type=inner_type):
            inner_gql_type = _parse_type_node(
                inner_type,
                schema=schema,
                not_null=False,
                context=context,
            )
            return GQLListType(type=inner_gql_type, not_null=not_null)
        case graphql.NonNullTypeNode(type=inner_type):
            return _parse_type_node(
                inner_type,
                schema=schema,
                not_null=True,
                context=context,
            )
        case _:
            msg = f"Unsupported type node: {type_node}"
            raise ValueError(msg)


def _parse_selection_set(
    selection_set: graphql.SelectionSetNode,
    parent_type: graphql.GraphQLCompositeType,
    ctx: ParseContext,
) -> list["GQLVar"]:
    return [
        v
        for sel in selection_set.selections
        for v in _parse_selection(sel, parent_type, ctx)
    ]


def _parse_output_type(
    out_type: graphql.GraphQLOutputType,
    ctx: ParseContext,
    *,
    not_null: bool = False,
    selection_set: graphql.SelectionSetNode | None = None,
) -> GQLType:
    match out_type:
        case (
            graphql.GraphQLObjectType()
            | graphql.GraphQLInterfaceType()
            | graphql.GraphQLUnionType()
        ):
            if not selection_set:
                type_kind = (
                    "union"
                    if isinstance(out_type, graphql.GraphQLUnionType)
                    else "object"
                )
                msg = f"Selection set is required for {type_kind} types"
                raise ValueError(msg)
            return GQLObjectType(
                type=out_type,
                selection=_parse_selection_set(selection_set, out_type, ctx),
                not_null=not_null,
            )
        case graphql.GraphQLNamedType():
            return GQLSingularType(type=out_type, not_null=not_null)
        case graphql.GraphQLNonNull():
            return _parse_output_type(
                out_type.of_type, ctx, not_null=True, selection_set=selection_set
            )
        case graphql.GraphQLList():
            inner_type = _parse_output_type(
                out_type.of_type, ctx, not_null=False, selection_set=selection_set
            )
            return GQLListType(type=inner_type, not_null=not_null)
        case _:
            msg = f"Unsupported output type: {out_type}"
            raise ValueError(msg)


def _parse_var(
    var_def: graphql.VariableDefinitionNode,
    *,
    schema: graphql.GraphQLSchema,
    context: str = "",
):
    var_name = var_def.variable.name.value
    var_context = _error_with_context(f"variable ${var_name}", context)
    gql_type = _parse_type_node(var_def.type, schema=schema, context=var_context)
    if var_def.default_value is not None:
        gql_type.default_value = value_from_ast_untyped(var_def.default_value)
    return GQLVar(
        name=var_name,
        type=gql_type,
        parent_type=None,
    )


def _parse_selection(
    selection: graphql.SelectionNode,
    parent_type: graphql.GraphQLCompositeType,
    ctx: ParseContext,
) -> list[GQLVar]:
    match selection:
        case graphql.FieldNode(name=name, alias=alias, selection_set=selection_set):
            field = schema_get_field(ctx.schema, parent_type, name.value)
            if not field:
                msg = f"Field '{name.value}' not found in type '{parent_type.name}'"
                raise ValueError(_error_with_context(msg, ctx.location))

            v = GQLVar(
                name=alias.value if alias else name.value,
                type=_parse_output_type(field.type, ctx, selection_set=selection_set),
                parent_type=parent_type,
            )
            return [v]
        case graphql.InlineFragmentNode(
            type_condition=tc, selection_set=selection_set
        ) if cast(graphql.NamedTypeNode | None, tc) is None:
            return _parse_selection_set(selection_set, parent_type, ctx)
        case graphql.InlineFragmentNode(
            type_condition=graphql.NamedTypeNode(
                name=graphql.NameNode(value=fragment_type_name)
            ),
            selection_set=selection_set,
        ):
            fragment_type = resolve_fragment_type(
                ctx.schema.get_type(fragment_type_name),
                fragment_type_name,
                parent_type,
            )
            return _parse_selection_set(selection_set, fragment_type, ctx)
        case graphql.FragmentSpreadNode(name=name):
            if name.value not in ctx.fragments:
                msg = f"Unknown fragment '{name.value}'"
                raise ValueError(_error_with_context(msg, ctx.location))
            fragment = ctx.fragments[name.value]
            fragment_type_name = fragment.type_condition.name.value
            fragment_type = resolve_fragment_type(
                ctx.schema.get_type(fragment_type_name),
                fragment_type_name,
                parent_type,
            )
            return _parse_selection_set(fragment.selection_set, fragment_type, ctx)
        case _:
            msg = f"Unsupported selection {selection} for parent type {parent_type}"
            raise ValueError(msg)


def resolve_fragment_type(
    fragment_type: graphql.GraphQLNamedType | None,
    fragment_type_name: str,
    parent_type: graphql.GraphQLCompositeType,
) -> graphql.GraphQLCompositeType:
    if not isinstance(
        fragment_type,
        (graphql.GraphQLObjectType, graphql.GraphQLInterfaceType),
    ):
        msg = f"Type condition '{fragment_type_name}' is not a composite type"
        raise TypeError(msg)
    if not isinstance(parent_type, graphql.GraphQLInterfaceType):
        return fragment_type
    if not implements_interface(fragment_type, parent_type):
        msg = (
            f"Type condition '{fragment_type.name}' does not "
            f"implement interface '{parent_type.name}'"
        )
        raise ValueError(msg)
    return fragment_type


def parse_input_type(
    input_type: graphql.GraphQLInputType,
    *,
    not_null: bool = False,
    default_value: Any = graphql.Undefined,
) -> GQLType:
    match input_type:
        case graphql.GraphQLNamedType():
            return GQLSingularType(
                type=input_type,
                not_null=not_null,
                default_value=default_value,
            )
        case graphql.GraphQLNonNull():
            return parse_input_type(
                input_type.of_type,
                not_null=True,
                default_value=default_value,
            )
        case graphql.GraphQLList():
            inner_type = parse_input_type(
                input_type.of_type,
                not_null=False,
                default_value=graphql.Undefined,
            )
            return GQLListType(
                type=inner_type,
                not_null=not_null,
                default_value=default_value,
            )
        case _:
            msg = f"Unsupported input type: {input_type}"
            raise ValueError(msg)


@dataclass(kw_only=True)
class ParseResult:
    queries: list[Query]
    error: str | None


def _parse_documents(
    statements: list[Statement],
) -> list[tuple[Statement, graphql.DocumentNode]]:
    return [(stmt, graphql.parse(stmt.clean_text)) for stmt in statements]


def _collect_fragments_from_doc(
    doc: graphql.DocumentNode,
) -> dict[str, graphql.FragmentDefinitionNode]:
    fragments: dict[str, graphql.FragmentDefinitionNode] = {}
    for definition in doc.definitions:
        if isinstance(definition, graphql.FragmentDefinitionNode):
            fragments[definition.name.value] = definition
    return fragments


def _collect_fragments(
    docs: list[tuple[Statement, graphql.DocumentNode]],
) -> dict[str, graphql.FragmentDefinitionNode]:
    fragments: dict[str, graphql.FragmentDefinitionNode] = {}
    for _, doc in docs:
        fragments.update(_collect_fragments_from_doc(doc))
    return fragments


def _collect_operations(
    docs: list[tuple[Statement, graphql.DocumentNode]],
) -> list[tuple[Statement, graphql.DocumentNode]]:
    operation_docs: list[tuple[Statement, graphql.DocumentNode]] = []
    for stmt, doc in docs:
        has_operations = any(
            isinstance(d, graphql.OperationDefinitionNode) for d in doc.definitions
        )
        if has_operations:
            operation_docs.append((stmt, doc))
    return operation_docs


def collect_fragment_spreads(node: graphql.Node) -> set[str]:
    spreads: set[str] = set()
    for child in node.keys:
        match getattr(node, child, None):
            case graphql.Node() as child_node:
                spreads.update(collect_fragment_spreads(child_node))
            case tuple() as items:
                for item in items:
                    match item:
                        case graphql.FragmentSpreadNode(name=name):
                            spreads.add(name.value)
                        case graphql.Node():
                            spreads.update(collect_fragment_spreads(item))
                        case _:
                            pass
            case _:
                pass
    return spreads


def _collect_referenced_fragment_names(
    doc: graphql.DocumentNode,
    fragments: dict[str, graphql.FragmentDefinitionNode],
) -> set[str]:
    def collect_fragment_names(name: str) -> set[str]:
        visited: set[str] = set()

        def walk(fragment_name: str) -> None:
            if fragment_name in visited:
                return
            fragment = fragments.get(fragment_name)
            if not fragment:
                return
            visited.add(fragment_name)
            for spread in collect_fragment_spreads(fragment):
                walk(spread)

        walk(name)
        return visited

    return {
        name
        for definition in doc.definitions
        if isinstance(definition, graphql.OperationDefinitionNode)
        for spread in collect_fragment_spreads(definition)
        for name in collect_fragment_names(spread)
    }


def _make_validation_doc(
    doc: graphql.DocumentNode,
    fragments: dict[str, graphql.FragmentDefinitionNode],
) -> graphql.DocumentNode:
    local_fragments = _collect_fragments_from_doc(doc)
    defined_fragments = set(local_fragments)
    effective_fragments = {**fragments, **local_fragments}
    referenced_fragments = _collect_referenced_fragment_names(doc, effective_fragments)
    extra_definitions = [
        effective_fragments[name]
        for name in sorted(referenced_fragments)
        if name not in defined_fragments
    ]
    return graphql.DocumentNode(definitions=[*doc.definitions, *extra_definitions])


def parse_gql_queries(
    schema_path: Path,
    statements: list[Statement],
    *,
    debug_path: Path | None = None,
) -> ParseResult:
    """Parse and validate GraphQL queries against a schema.

    Args:
        schema_path: Path to GraphQL schema file (SDL format)
        statements: List of GraphQL query statements to parse
        debug_path: Optional directory to save debug artifacts (schema, queries, AST)

    Returns:
        ParseResult containing validated queries or validation errors
    """
    schema_document = graphql.parse(schema_path.read_text(encoding="utf-8"))
    schema = graphql.build_ast_schema(schema_document)

    docs = _parse_documents(statements)
    fragments = _collect_fragments(docs)
    operation_docs = _collect_operations(docs)

    queries = []
    for stmt, doc in operation_docs:
        validation_doc = _make_validation_doc(doc, fragments)
        exec_source = graphql.print_ast(validation_doc)
        queries.append(
            Query(
                stmt=stmt,
                doc=validation_doc,
                schema=schema,
                fragments=_collect_fragments_from_doc(validation_doc),
                exec_source=exec_source,
            )
        )

    errors = []
    for q in queries:
        errs = graphql.validate(q.schema, q.doc)
        if errs:
            errors.append(
                f"Invalid GraphQL query in {q.stmt.location}:\n"
                + "\n".join(str(e) for e in errs)
            )
            continue
        try:
            _ = q.selection_set
        except (ValueError, TypeError) as exc:
            errors.append(f"Invalid GraphQL query in {q.stmt.location}:\n{exc}")

    if debug_path:
        debug_path.mkdir(parents=True, exist_ok=True)
        shutil.copy(schema_path, debug_path / "schema.graphql")
        _dump_strings(debug_path / "queries.gql", [q.stmt.clean_text for q in queries])
        _dump_json(debug_path / "queries.json", [q.doc.to_dict() for q in queries])
        _dump_json(debug_path / "schema.json", schema_document.to_dict())
        _dump_json(
            debug_path / "out.json",
            [
                {
                    "stmt": q.stmt.clean_text,
                    "location": q.stmt.location,
                    "name": q.name,
                    "variables": q.variables,
                    "selection_set": q.selection_set,
                }
                for q in queries
            ],
        )

    return ParseResult(queries=queries, error="\n".join(errors) if errors else None)


def _dump_json(path: Path, obj: object):
    path.write_bytes(
        pydantic.TypeAdapter(type(obj)).dump_json(obj, indent=2, fallback=str)
    )


def _dump_strings(path: Path, strings: list[str]):
    path.write_text("\n\n".join(strings), encoding="utf-8")


# Port of GraphQLSchema.get_field from graphql-core 3.3
# See https://github.com/graphql-python/graphql-core/blob/main/src/graphql/type/schema.py#L374
def schema_get_field(
    schema: graphql.GraphQLSchema,
    parent_type: graphql.GraphQLCompositeType,
    field_name: str,
) -> graphql.GraphQLField | None:
    if field_name == "__schema":
        return graphql.SchemaMetaFieldDef if schema.query_type is parent_type else None
    if field_name == "__type":
        return graphql.TypeMetaFieldDef if schema.query_type is parent_type else None
    if field_name == "__typename":
        return graphql.TypeNameMetaFieldDef

    try:
        # This is a port not reimplementation, so we use author's approach.
        return parent_type.fields[field_name]  # pyright: ignore[reportAttributeAccessIssue]
    except (AttributeError, KeyError):
        return None


def get_transitive_interfaces(
    type_: graphql.GraphQLObjectType | graphql.GraphQLInterfaceType,
) -> set[graphql.GraphQLInterfaceType]:
    interfaces = set()
    queue = list(type_.interfaces)
    while queue:
        current = queue.pop()
        if current in interfaces:
            continue
        interfaces.add(current)
        queue.extend(current.interfaces)
    return interfaces


def implements_interface(
    type_: graphql.GraphQLObjectType | graphql.GraphQLInterfaceType,
    interface: graphql.GraphQLInterfaceType,
) -> bool:
    return type_ == interface or interface in get_transitive_interfaces(type_)



import ast
import logging
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from pathlib import Path

import graphql
from pydantic import alias_generators

from iron_gql.parser import GQLListType
from iron_gql.parser import GQLObjectType
from iron_gql.parser import GQLSingularType
from iron_gql.parser import GQLType
from iron_gql.parser import GQLVar
from iron_gql.parser import Query
from iron_gql.parser import Statement
from iron_gql.parser import get_transitive_interfaces
from iron_gql.parser import parse_gql_queries
from iron_gql.parser import parse_input_type
from iron_gql.util import capitalize_first

logger = logging.getLogger(__name__)

type StrTransform = Callable[[str], str]

BUILTIN_SCALARS = {
    "String": "str",
    "Int": "int",
    "Float": "float",
    "Boolean": "bool",
    "Date": "datetime.date",
    "DateTime": "datetime.datetime",
    "JSON": "object",
    "Upload": "runtime.FileVar",
}


@dataclass(kw_only=True)
class CollectorContext:
    enums: set[graphql.GraphQLEnumType] = dataclass_field(default_factory=set)


@dataclass(kw_only=True)
class GeneratedModel:
    name: str
    fields: list[str]


@dataclass(kw_only=True)
class RenderedFieldModels:
    child_models: list[GeneratedModel] = dataclass_field(default_factory=list)
    child_model_name_base: str | None = None
    union_types: list[str] = dataclass_field(default_factory=list)
    union_field_type: GQLType | None = None


def finalize_type(
    typ: str,
    field_typ: GQLType,
    *,
    include_default: bool = True,
) -> str:
    if not field_typ.not_null:
        typ += " | None"
    if include_default and field_typ.default_value != graphql.Undefined:
        typ += f" = {field_typ.default_value!r}"
    return typ


def _format_discriminated_union_type(
    branch_types: list[str],
    field_type: GQLType,
) -> str:
    match field_type:
        case GQLListType(type=inner_type):
            inner = _format_discriminated_union_type(branch_types, inner_type)
            typ = f"list[{inner}]"
        case _:
            union_str = " | ".join(branch_types)
            typ = f'Annotated[{union_str}, pydantic.Field(discriminator="typename__")]'
    return finalize_type(typ, field_type)


def unwrap_input_type(
    input_type: graphql.GraphQLInputType,
) -> graphql.GraphQLInputType:
    while isinstance(input_type, (graphql.GraphQLNonNull, graphql.GraphQLList)):
        input_type = input_type.of_type
    return input_type


def render_pydantic_class(name: str, base: str, fields: list[str]) -> str:
    return f"class {name}({base}):\n    {indent_block(chr(10).join(fields), '    ')}"


def generate_gql_package(
    *,
    schema_path: Path,
    package_full_name: str,
    base_url_import: str,
    scalars: dict[str, str] | None = None,
    to_camel_fn_full_name: str = "pydantic.alias_generators:to_camel",
    to_snake_fn: StrTransform = alias_generators.to_snake,
    debug_path: Path | None = None,
    src_path: Path,
) -> bool:
    """Generate a typed GraphQL client from schema and discovered queries.

    Scans src_path for calls to `<package>_gql()`, validates queries against
    schema_path, and generates a module with Pydantic models and typed query
    classes with async execution methods.

    Args:
        schema_path: Path to GraphQL SDL schema file
        package_full_name: Full module name for generated package
            (e.g., "myapp.gql.client")
        base_url_import: Import path to base URL
            (e.g., "myapp.config:GRAPHQL_URL")
        scalars: Custom GraphQL scalar to Python type mapping
            (e.g., {"ID": "builtins:str"})
        to_camel_fn_full_name: Import path to camelCase conversion function
        to_snake_fn: Function for converting names to snake_case
        debug_path: Optional path for saving debug artifacts
        src_path: Root directory to search for GraphQL query calls

    Returns:
        True if the generated file was modified, False if content unchanged
    """
    if scalars is None:
        scalars = {}

    package_name = package_full_name.split(".")[-1]  # noqa: PLC0207
    gql_fn_name = f"{package_name}_gql"

    target_package_path = src_path / f"{package_full_name.replace('.', '/')}.py"
    base_url_import_package, base_url_import_path = base_url_import.split(":")

    queries = list(
        find_all_queries(src_path, gql_fn_name, skip_path=target_package_path)
    )

    parse_res = parse_gql_queries(
        schema_path,
        queries,
        debug_path=debug_path,
    )

    if parse_res.error:
        logger.error(parse_res.error)
        return False

    schema_base = schema_path.resolve()
    src_base = src_path.resolve()
    schema_for_render = schema_base.relative_to(src_base, walk_up=True)

    new_content = render_package(
        base_url_import_package=base_url_import_package,
        base_url_import_path=base_url_import_path,
        schema_path=schema_for_render,
        package_name=package_name,
        gql_fn_name=gql_fn_name,
        queries=sorted(parse_res.queries, key=lambda q: q.name),
        scalars=scalars,
        to_camel_fn_full_name=to_camel_fn_full_name,
        to_snake_fn=to_snake_fn,
    )
    changed = write_if_changed(target_package_path, new_content + "\n")
    if changed:
        logger.info(f"Generated GQL package {package_full_name}")
    return changed


def find_fn_calls(
    root_path: Path, fn_name: str, *, skip_path: Path
) -> Iterator[tuple[Path, int, ast.Call]]:
    for path in root_path.glob("**/*.py"):
        if path.resolve() == skip_path.resolve():
            continue
        content = path.read_text(encoding="utf-8")
        if fn_name not in content:
            continue
        for node in ast.walk(ast.parse(content, filename=str(path))):
            match node:
                case ast.Call(func=ast.Name(id=id)) if id == fn_name:
                    yield path, node.lineno, node
                case _:
                    pass


def find_all_queries(
    src_path: Path, gql_fn_name: str, *, skip_path: Path
) -> Iterator[Statement]:
    for file, lineno, node in find_fn_calls(src_path, gql_fn_name, skip_path=skip_path):
        relative_path = file.relative_to(src_path)

        stmt_arg = node.args[0]
        if (
            len(node.args) != 1
            or not isinstance(stmt_arg, ast.Constant)
            or not isinstance(stmt_arg.value, str)
        ):
            msg = (
                f"Invalid positional arguments for {gql_fn_name} "
                f"at {relative_path}:{lineno}, "
                "expected a single string literal"
            )
            raise TypeError(msg)

        yield Statement(raw_text=stmt_arg.value, file=relative_path, lineno=lineno)


HEADER = """\
# Code generated by iron_gql, DO NOT EDIT.
# fmt: off
# pyright: reportUnusedImport=false
# ruff: noqa"""


def _render_imports(
    to_camel_fn_full_name: str,
    scalars: dict[str, str],
    base_url_import_package: str,
    base_url_import_path: str,
) -> str:
    import_modules = [m.split(":")[0] for m in scalars.values()]
    scalar_imports = "\n".join(f"import {m}" for m in import_modules)
    to_camel_module = to_camel_fn_full_name.split(":", maxsplit=1)[0]
    base_url_symbol = base_url_import_path.split(".", maxsplit=1)[0]
    return f"""\
import datetime
from pathlib import Path
from typing import Annotated
from typing import Literal
from typing import overload

import pydantic

from iron_gql import runtime

import {to_camel_module}

{scalar_imports}

from {base_url_import_package} import {base_url_symbol}"""


def _render_client_init(
    package_name: str,
    base_url_import_path: str,
    schema_path: Path,
    to_camel_fn_full_name: str,
) -> str:
    return f"""\
{package_name.upper()}_CLIENT = runtime.GQLClient(
    base_url={base_url_import_path},
    schema=Path("{schema_path}"),
)


class GQLModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        populate_by_name=True,
        alias_generator={to_camel_fn_full_name.replace(":", ".")},
        extra="forbid",
        validate_default=True,
    )"""


def _render_gql_fn(
    queries: list[Query],
    package_name: str,
    gql_fn_name: str,
) -> str:
    dict_name = f"_{package_name.upper()}_GQL_DISPATCH"

    entries = [(repr(q.stmt.raw_text), capitalize_first(q.name)) for q in queries]

    overloads = "\n".join(
        f"@overload\ndef {gql_fn_name}(stmt: Literal[{lit}]) -> {ret}: ..."
        for lit, ret in entries
    )
    dict_entries = "\n".join(f"    {lit}: {ret}," for lit, ret in entries)

    return f"""\
{overloads}
@overload
def {gql_fn_name}(stmt: str) -> runtime.GQLQuery: ...


{dict_name}: dict[str, type[runtime.GQLQuery]] = {{
{dict_entries}
}}


def {gql_fn_name}(stmt: str) -> runtime.GQLQuery:
    query_cls = {dict_name}.get(stmt)
    if query_cls is not None:
        return query_cls()
    return runtime.GQLQuery()"""


def render_package(
    base_url_import_package: str,
    base_url_import_path: str,
    schema_path: Path,
    package_name: str,
    gql_fn_name: str,
    queries: list[Query],
    scalars: dict[str, str],
    to_camel_fn_full_name: str,
    to_snake_fn: StrTransform,
):
    queries = get_unique_queries(queries)
    ctx = CollectorContext()

    # Collect input type roots from all query variables
    input_roots: set[graphql.GraphQLInputObjectType] = set()
    for query in queries:
        for v in query.variables:
            _collect_input_root(v.type, input_roots)
    ordered_input_types = collect_ordered_input_types(input_roots)

    rendered_query_classes = render_query_classes(
        queries, package_name, scalars, to_snake_fn, ctx
    )
    rendered_result_models = render_result_models(queries, scalars, to_snake_fn, ctx)
    rendered_input_types = render_input_types(
        ordered_input_types,
        scalars,
        to_snake_fn,
        ctx,
    )
    rendered_enums = render_enums(ctx.enums)

    sections = [
        HEADER,
        _render_imports(
            to_camel_fn_full_name,
            scalars,
            base_url_import_package,
            base_url_import_path,
        ),
        _render_client_init(
            package_name, base_url_import_path, schema_path, to_camel_fn_full_name
        ),
        "\n".join(rendered_enums),
        "\n\n\n".join(rendered_result_models),
        "\n\n\n".join(rendered_input_types),
        "\n\n\n".join(rendered_query_classes),
        _render_gql_fn(queries, package_name, gql_fn_name),
    ]
    return "\n\n\n".join(section for section in sections if section)


def _collect_input_root(
    gql_type: GQLType,
    roots: set[graphql.GraphQLInputObjectType],
) -> None:
    match gql_type:
        case GQLSingularType(type=typ) if isinstance(
            typ, graphql.GraphQLInputObjectType
        ):
            roots.add(typ)
        case GQLListType(type=inner):
            _collect_input_root(inner, roots)
        case _:
            pass


def get_unique_queries(queries: list[Query]) -> list[Query]:
    unique_queries: dict[str, Query] = {}
    for query in queries:
        if query.name not in unique_queries:
            unique_queries[query.name] = query
        elif unique_queries[query.name].stmt.hash_str != query.stmt.hash_str:
            msg = (
                f"Cannot compile different GraphQL queries with same name {query.name}"
                f" at {query.stmt.location}"
                f" and {unique_queries[query.name].stmt.location}"
            )
            raise ValueError(msg)
    return list(unique_queries.values())


def render_enums(enum_types: set[graphql.GraphQLEnumType]) -> list[str]:
    return [
        f"type {typ.name} = Literal[{', '.join(repr(name) for name in typ.values)}]"
        for typ in sorted(enum_types, key=lambda t: t.name)
    ]


def _extract_object_selection(
    field_type: GQLType,
) -> tuple[graphql.GraphQLNamedType, list[GQLVar]] | None:
    match field_type:
        case GQLObjectType(type=obj_type, selection=selection):
            return obj_type, selection
        case GQLListType(type=GQLObjectType(type=obj_type, selection=selection)):
            return obj_type, selection
        case _:
            return None


@dataclass(kw_only=True)
class ResultModelRenderer:
    scalars: dict[str, str]
    to_snake_fn: StrTransform
    ctx: CollectorContext
    schema: graphql.GraphQLSchema

    def render_models(
        self,
        model_name_base: str,
        fields: list[GQLVar],
        model_type: graphql.GraphQLNamedType,
        *,
        typename_type: str | None = None,
    ) -> list[GeneratedModel]:
        child_models: list[GeneratedModel] = []
        fields_mapping: dict[str, str] = {}
        for field in fields:
            rendered = self._render_field_models(model_name_base, field)
            child_models.extend(rendered.child_models)

            if field.name.startswith("__"):
                py_name = self.to_snake_fn(f"{field.name[2:]}__")
                py_alias = field.name
            else:
                py_name = self.to_snake_fn(field.name)
                py_alias = None

            py_type = self._render_field_type(
                field,
                rendered,
                model_type,
                typename_type,
            )
            if py_alias:
                py_type = f'{py_type} = pydantic.Field(alias="{field.name}")'
            fields_mapping[py_name] = py_type
        return [
            *child_models,
            GeneratedModel(
                name=model_name_base,
                fields=[
                    f"{field}: {field_type}"
                    for field, field_type in fields_mapping.items()
                ],
            ),
        ]

    def _render_field_models(
        self,
        model_name_base: str,
        field: GQLVar,
    ) -> RenderedFieldModels:
        obj_info = _extract_object_selection(field.type)
        if obj_info is None:
            return RenderedFieldModels()
        obj_type, selection = obj_info
        match obj_type:
            case graphql.GraphQLObjectType():
                return self._render_object_field(
                    model_name_base,
                    field.name,
                    obj_type,
                    selection,
                )
            case graphql.GraphQLUnionType():
                return self._render_union_field(
                    model_name_base,
                    field.name,
                    obj_type,
                    selection,
                    field.type,
                )
            case graphql.GraphQLInterfaceType():
                return self._render_interface_field(
                    model_name_base,
                    field.name,
                    obj_type,
                    selection,
                    field.type,
                )
            case _:
                msg = f"Unknown type {obj_type} for field {field.name}"
                raise ValueError(msg)

    def _render_object_field(
        self,
        model_name_base: str,
        field_name: str,
        obj_type: graphql.GraphQLObjectType,
        selection: list[GQLVar],
    ) -> RenderedFieldModels:
        child_model_name_base = model_name_base + capitalize_first(field_name)
        child_models = self.render_models(child_model_name_base, selection, obj_type)
        return RenderedFieldModels(
            child_models=child_models,
            child_model_name_base=child_model_name_base,
            union_types=[],
            union_field_type=None,
        )

    def _render_union_field(
        self,
        model_name_base: str,
        field_name: str,
        union_type: graphql.GraphQLUnionType,
        selection: list[GQLVar],
        field_type: GQLType,
    ) -> RenderedFieldModels:
        self._require_union_typename(selection, union_type)
        child_models: list[GeneratedModel] = []
        union_types: list[str] = []
        for subtyp in union_type.types:
            eligible_parents = {union_type, subtyp} | get_transitive_interfaces(subtyp)
            subtyp_sel = [
                sel_field
                for sel_field in selection
                if sel_field.parent_type in eligible_parents
            ]
            child_model_name_base = (
                model_name_base + capitalize_first(field_name) + subtyp.name
            )
            union_types.append(child_model_name_base)
            child_models.extend(
                self.render_models(child_model_name_base, subtyp_sel, subtyp)
            )
        return RenderedFieldModels(
            child_models=child_models,
            child_model_name_base=None,
            union_types=union_types,
            union_field_type=field_type,
        )

    def _render_interface_field(
        self,
        model_name_base: str,
        field_name: str,
        interface_type: graphql.GraphQLInterfaceType,
        selection: list[GQLVar],
        field_type: GQLType,
    ) -> RenderedFieldModels:
        explicit_fragment_types = self._collect_explicit_fragment_types(
            selection, interface_type
        )
        if not explicit_fragment_types:
            child_model_name_base = model_name_base + capitalize_first(field_name)
            child_models = self.render_models(
                child_model_name_base,
                selection,
                interface_type,
                typename_type="str",
            )
            return RenderedFieldModels(
                child_models=child_models,
                child_model_name_base=child_model_name_base,
                union_types=[],
                union_field_type=None,
            )

        self._require_interface_typename(selection, interface_type)
        explicit_objects = self._expand_fragment_objects(explicit_fragment_types)
        fallback_objects = (
            self._possible_interface_objects(interface_type) - explicit_objects
        )
        interface_model_base = model_name_base + capitalize_first(field_name)
        child_models, union_types = self._render_interface_union_models(
            interface_model_base,
            interface_type,
            explicit_objects,
            fallback_objects,
            selection,
        )
        return RenderedFieldModels(
            child_models=child_models,
            child_model_name_base=None,
            union_types=union_types,
            union_field_type=field_type,
        )

    def _collect_explicit_fragment_types(
        self,
        selection: list[GQLVar],
        interface_type: graphql.GraphQLInterfaceType,
    ) -> set[graphql.GraphQLNamedType]:
        return {
            sel_field.parent_type
            for sel_field in selection
            if sel_field.parent_type is not None
            and sel_field.parent_type != interface_type
        }

    def _require_interface_typename(
        self,
        selection: list[GQLVar],
        interface_type: graphql.GraphQLInterfaceType,
    ) -> None:
        has_typename = any(
            sel_field.name == "__typename" and sel_field.parent_type == interface_type
            for sel_field in selection
        )
        if not has_typename:
            msg = (
                "Missing __typename in selection set for interface "
                f"'{interface_type.name}'"
            )
            raise ValueError(msg)

    def _require_union_typename(
        self,
        selection: list[GQLVar],
        union_type: graphql.GraphQLUnionType,
    ) -> None:
        typename_parents = {
            sel_field.parent_type
            for sel_field in selection
            if sel_field.name == "__typename" and sel_field.parent_type is not None
        }
        missing: list[str] = []
        for subtyp in union_type.types:
            eligible_parents = {union_type, subtyp} | get_transitive_interfaces(subtyp)
            if not (eligible_parents & typename_parents):
                missing.append(subtyp.name)
        if missing:
            msg = f"Missing __typename in selection set for union '{union_type.name}'"
            raise ValueError(msg)

    def _possible_interface_objects(
        self,
        interface_type: graphql.GraphQLInterfaceType,
    ) -> set[graphql.GraphQLObjectType]:
        return set(self.schema.get_possible_types(interface_type))

    def _expand_fragment_objects(
        self,
        fragment_types: set[graphql.GraphQLNamedType],
    ) -> set[graphql.GraphQLObjectType]:
        explicit_objects: set[graphql.GraphQLObjectType] = set()
        for fragment_type in fragment_types:
            if isinstance(fragment_type, graphql.GraphQLObjectType):
                explicit_objects.add(fragment_type)
            elif isinstance(fragment_type, graphql.GraphQLInterfaceType):
                explicit_objects.update(self._possible_interface_objects(fragment_type))
        return explicit_objects

    def _render_interface_union_models(
        self,
        interface_model_base: str,
        interface_type: graphql.GraphQLInterfaceType,
        explicit_objects: set[graphql.GraphQLObjectType],
        fallback_objects: set[graphql.GraphQLObjectType],
        selection: list[GQLVar],
    ) -> tuple[list[GeneratedModel], list[str]]:
        child_models: list[GeneratedModel] = []
        union_types: list[str] = []
        for obj_type in sorted(explicit_objects, key=lambda t: t.name):
            child_model_name_base = interface_model_base + obj_type.name
            obj_interfaces = get_transitive_interfaces(obj_type)
            eligible_parents = {interface_type, obj_type} | obj_interfaces
            obj_sel = [
                sel_field
                for sel_field in selection
                if sel_field.parent_type in eligible_parents
            ]
            child_models.extend(
                self.render_models(child_model_name_base, obj_sel, obj_type)
            )
            union_types.append(child_model_name_base)

        if fallback_objects:
            fallback_name = interface_model_base + interface_type.name
            fallback_sel = [
                sel_field
                for sel_field in selection
                if sel_field.parent_type == interface_type
            ]
            fallback_type_names = sorted(obj.name for obj in fallback_objects)
            fallback_typename_literal = (
                f"Literal[{', '.join(repr(name) for name in fallback_type_names)}]"
            )
            child_models.extend(
                self.render_models(
                    fallback_name,
                    fallback_sel,
                    interface_type,
                    typename_type=fallback_typename_literal,
                )
            )
            union_types.append(fallback_name)
        return child_models, union_types

    def _render_field_type(
        self,
        field: GQLVar,
        rendered: RenderedFieldModels,
        model_type: graphql.GraphQLNamedType,
        typename_type: str | None,
    ) -> str:
        if field.name == "__typename":
            return typename_type or f'Literal["{model_type.name}"]'
        if rendered.union_types:
            if rendered.union_field_type is None:
                msg = "Union field type is required for union rendering"
                raise ValueError(msg)
            return _format_discriminated_union_type(
                rendered.union_types, rendered.union_field_type
            )
        return field_type(
            field.type,
            self.scalars,
            child_model_name_base=rendered.child_model_name_base,
            ctx=self.ctx,
        )


def render_result_models(
    queries: list[Query],
    scalars: dict[str, str],
    to_snake_fn: StrTransform,
    ctx: CollectorContext,
):
    if not queries:
        return []
    renderer = ResultModelRenderer(
        scalars=scalars,
        to_snake_fn=to_snake_fn,
        ctx=ctx,
        schema=queries[0].schema,
    )

    return [
        render_pydantic_class(model.name, "GQLModel", model.fields)
        for query in queries
        for model in renderer.render_models(
            f"{capitalize_first(query.name)}Result",
            query.selection_set,
            query.root_type,
        )
    ]


def render_query_classes(
    queries: list[Query],
    package_name: str,
    scalars: dict[str, str],
    to_snake_fn: StrTransform,
    ctx: CollectorContext,
) -> list[str]:
    query_classes = []
    for query in queries:
        args = ["self"]
        variables = []
        if query.variables:
            args.append("*")
        for v in query.variables:
            py_name = to_snake_fn(v.name)
            typ = field_type(v.type, scalars, ctx=ctx)
            args.append(f"{py_name}: {typ}")
            variables.append(f'"{v.name}": runtime.serialize_var({py_name})')
        variables_arg = (
            f"\n            {{{', '.join(variables)}}}," if variables else ""
        )

        query_classes.append(
            f"""

class {capitalize_first(query.name)}(runtime.GQLQuery):
    async def execute({", ".join(args)}) -> {capitalize_first(query.name)}Result:
        return await {package_name.upper()}_CLIENT.query(
            {capitalize_first(query.name)}Result,
            {query.exec_source!r},{variables_arg}
            headers=self.headers,
            upload_files=self.upload_files,
        )

            """.strip()
        )
    return query_classes


def render_input_types(
    ordered_input_types: list[graphql.GraphQLInputObjectType],
    scalars: dict[str, str],
    to_snake_fn: StrTransform,
    ctx: CollectorContext | None = None,
) -> list[str]:
    rendered: list[str] = []
    for typ in ordered_input_types:
        fields = [
            f"{to_snake_fn(field_name)}: {
                input_field_type(
                    parse_input_type(field.type, default_value=field.default_value),
                    scalars,
                    ctx=ctx,
                )
            }"
            for field_name, field in typ.fields.items()
        ]
        rendered.append(render_pydantic_class(typ.name, "GQLModel", fields))
    return rendered


def collect_ordered_input_types(
    roots: Iterable[graphql.GraphQLInputObjectType],
) -> list[graphql.GraphQLInputObjectType]:
    emitted: set[str] = set()
    ordered: list[graphql.GraphQLInputObjectType] = []

    def emit(typ: graphql.GraphQLInputObjectType) -> None:
        if typ.name in emitted:
            return
        emitted.add(typ.name)
        for field in typ.fields.values():
            target = unwrap_input_type(field.type)
            if isinstance(target, graphql.GraphQLInputObjectType):
                emit(target)
        ordered.append(typ)

    for root in sorted(roots, key=lambda t: t.name):
        emit(root)

    return ordered


def input_field_type(
    field_typ: GQLType,
    scalars: dict[str, str],
    *,
    ctx: CollectorContext | None = None,
) -> str:
    has_input_object, typ = _build_input_type(field_typ, scalars, ctx=ctx)
    if has_input_object:
        typ = f'"{typ}"'
    if field_typ.default_value != graphql.Undefined:
        typ += f" = {field_typ.default_value!r}"
    elif not field_typ.not_null:
        typ += " = None"
    return typ


def _build_input_type(
    field_typ: GQLType,
    scalars: dict[str, str],
    *,
    ctx: CollectorContext | None = None,
) -> tuple[bool, str]:
    match field_typ:
        case GQLSingularType(type=gql_type):
            has_input, base_typ = _input_py_type(gql_type, scalars, ctx=ctx)
        case GQLListType(type=inner):
            has_input, child_type = _build_input_type(inner, scalars, ctx=ctx)
            base_typ = f"list[{child_type}]"
        case _:
            msg = f"Unknown GQLType {field_typ} of type {type(field_typ)}"
            raise TypeError(msg)
    return has_input, finalize_type(base_typ, field_typ, include_default=False)


def _input_py_type(
    gql_type: graphql.GraphQLNamedType,
    scalars: dict[str, str],
    *,
    ctx: CollectorContext | None = None,
) -> tuple[bool, str]:
    match gql_type:
        case graphql.GraphQLScalarType(name=name):
            if name in scalars:
                return False, scalars[name].replace(":", ".")
            if name in BUILTIN_SCALARS:
                return False, BUILTIN_SCALARS[name]
            logger.warning(f"Unknown scalar type {name}")
            return False, "object"
        case graphql.GraphQLInputObjectType(name=name):
            return True, name
        case graphql.GraphQLEnumType(name=name):
            if ctx is not None:
                ctx.enums.add(gql_type)
            return False, name
        case _:
            logger.warning(f"Unknown GraphQL type {gql_type.name} {type(gql_type)}")
            return False, "object"


def field_type(
    field_typ: GQLType,
    scalars: dict[str, str],
    *,
    child_model_name_base: str | None = None,
    ctx: CollectorContext | None = None,
) -> str:
    match field_typ:
        case GQLSingularType():
            typ = field_py_type(field_typ.type, scalars, ctx=ctx)
        case GQLObjectType():
            if child_model_name_base is None:
                msg = "child_model_name_base must be provided for GQLObjectType"
                raise ValueError(msg)
            typ = child_model_name_base
        case GQLListType():
            child_type = field_type(
                field_typ.type,
                scalars,
                child_model_name_base=child_model_name_base,
                ctx=ctx,
            )
            typ = f"list[{child_type}]"
        case _:
            msg = f"Unknown GQLType {field_typ} of type {type(field_typ)}"
            raise TypeError(msg)
    return finalize_type(typ, field_typ)


def field_py_type(
    gql_type: graphql.GraphQLNamedType,
    scalars: dict[str, str],
    *,
    ctx: CollectorContext | None = None,
) -> str:
    match gql_type:
        case graphql.GraphQLScalarType(name=name):
            if name in scalars:
                return scalars[name].replace(":", ".")
            if name in BUILTIN_SCALARS:
                return BUILTIN_SCALARS[name]
            logger.warning(f"Unknown scalar type {name}")
            return "object"
        case graphql.GraphQLInputObjectType(name=name):
            return name
        case graphql.GraphQLEnumType(name=name):
            if ctx is not None:
                ctx.enums.add(gql_type)
            return name
        case _:
            logger.warning(f"Unknown GraphQL type {gql_type.name} {type(gql_type)}")
            return "object"


def indent_block(block: str, indent: str) -> str:
    return "\n".join(
        indent + line if i > 0 and line.strip() else line
        for i, line in enumerate(block.split("\n"))
    )


def write_if_changed(path: Path, new_content: str) -> bool:
    path.parent.mkdir(parents=True, exist_ok=True)
    existing_content = path.read_text(encoding="utf-8") if path.exists() else None
    if existing_content == new_content:
        return False
    path.write_text(new_content, encoding="utf-8")
    path.touch()
    return True
